import os
import csv
import time
import logging
import subprocess
from datetime import datetime

from ruxit.api.base_plugin import BasePlugin
from ruxit.api.selectors import HostSelector

from run_as import run_as
from status import schtask_success_statuses, schtask_failed_statuses


logger = logging.getLogger(__name__)


class Tasks_Class(BasePlugin):
    def query(self, **kwargs):
        logger.setLevel(self.config.get("log_level", "INFO"))
        self.domain = "."
        self.username = self.config.get("username", "")
        if self.username:
            if "\\" in self.username:
                self.domain, self.username = self.username.split("\\")
        self.password = self.config.get("password", "")
        self.manual_mode = self.config.get("manual_mode", False)
        self.time_format = self.config.get("time_format", "%d/%m/%Y %H:%M:%S")
        if self.manual_mode:
            self.data_folder = self.config.get("folder")

        tasks = self.config["tasks"].strip().split('\n')
        tasks = [(name.strip(), int(timeout.strip())) for name, timeout in list(t.split(":") for t in tasks)]

        if self.manual_mode:
            state_output, details_output = self.get_data_from_file(os.path.abspath(self.data_folder))
        else:
            state_output, details_output = self.get_bulk_tasks()
        
        if state_output:
            logger.debug("Raw state output:\n")
            logger.debug(state_output)
            task_states = csv.reader(state_output.splitlines()[2:])
            task_states = {ts[0]: ts[1] for ts in task_states for name, _ in tasks if name.lower() in ts[0].lower()}
            logger.debug("Task states:\n")
            logger.debug(task_states)
            for name, state in task_states.items():
                self.results_builder.state_metric("task_state", state, dimensions={"Task": name})
       
        if details_output:
            logger.debug("Raw details output:")
            logger.debug(details_output)
            for task_name, timeout in tasks:
                task_details = csv.reader(details_output.splitlines()[2:])
                logger.info(f"Getting additional details for task {task_name}")
                line = [(td[0], td[1], td[2]) for td in task_details if task_name.lower() in td[2].lower()]
                line = line[0] if line else None
                if line:
                    logger.info(f"Found additional details as: {line}")
                    try:
                        name = line[2]
                        state = task_states.get(name, "")
                        last_run = datetime.strptime(line[0], self.time_format).astimezone().timestamp()
                        time_since_run = int((datetime.now().astimezone().timestamp() - last_run)*1000)
                        last_result = str(line[1])
                    except:
                        self.results_builder.report_custom_info_event(
                            title=f"Cannot read task details for task {task_name}",
                            description=f"Line: {line}",
                            entity_selector=HostSelector()
                        )
                        continue

                    if last_run:
                        self.results_builder.absolute(key="time_since_exec", value=time_since_run, dimensions={"Task": name})
                    
                    # Check for fail states and create problems
                    if state == 'Running' and time_since_run >= timeout*60_000:
                        self.results_builder.report_error_event(
                            f'Task "{name}" appears to be hanging. It is still running after {timeout} minutes', 
                            f'Task "{name}" appears to be hanging',
                            entity_selector=HostSelector()
                        )
                        logger.info(f'{name} appears stuck. It is in {state} state. It did not finish in {timeout} minutes')
                    elif last_result not in schtask_success_statuses:
                        self.results_builder.report_error_event(
                            f'Task "{name}" did not finish successfully. Last result was: {last_result} (0x{int(last_result):x}); \
                                {schtask_failed_statuses.get(last_result, "")}',
                            f'Task "{name}" failed',
                            entity_selector=HostSelector()
                        )
                        logger.info(
                            f'{name} did not finish successfully. Last result was {last_result} (0x{int(last_result):x}); {schtask_failed_statuses.get(last_result, "")}'
                        )

    def get_bulk_tasks(self):
        state_output = ""
        details_output = ""

        if self.username and self.password:
            logger.info(f"Running as user {self.username}")
            temp_folder = os.path.join(os.environ["systemroot"], "Temp")
            run_as(self.username, self.password, 'powershell.exe "Get-ScheduledTask | Select TaskName, State | ConvertTo-Csv" > $env:systemroot\Temp\scheduledtasks.csv', self.domain)
            run_as(self.username, self.password, 'powershell.exe "Get-ScheduledTask | Get-ScheduledTaskInfo | Select LastRunTime, LastTaskResult, TaskName | ConvertTo-Csv" > $env:systemroot\Temp\scheduledtasksinfo.csv', self.domain)
            time.sleep(5) # run_as is async. making sure cmd finished
            state_output, details_output = self.get_data_from_file(temp_folder)
        else:
            logger.info(f"Running powershell commands to get bulk task info")
            state_stdout, state_stderr = subprocess.Popen(
                args=['powershell.exe', "Get-ScheduledTask | Select TaskName, State | ConvertTo-Csv"],
                shell=True,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            ).communicate()
            if state_stderr:
                logger.error("Error when running Get-ScheduledTask: " + state_stderr.decode())
                return "", ""
            details_stdout, details_stderr = subprocess.Popen(
                args=['powershell.exe', "Get-ScheduledTask | Get-ScheduledTaskInfo | Select LastRunTime, LastTaskResult, TaskName | ConvertTo-Csv"],
                shell=True,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            ).communicate()
            if details_stderr:
                logger.error("Error when running Get-ScheduledTaskInfo: " + details_stderr.decode())
                return "", ""
            try:
                state_output = state_stdout.decode()
                details_output = details_stdout.decode()
            except UnicodeEncodeError:
                try:
                    state_output = state_stdout.decode(encoding="utf-16")
                    details_output = details_stdout.decode(encoding="utf-16")
                except UnicodeDecodeError:
                    logger.error("Could not decode powershell commands output")
                    return "", ""

        return state_output, details_output

    def get_data_from_file(self, folder):
        state_output, details_output = "", ""
        state_file = os.path.join(folder, "scheduledtasks.csv")
        details_file = os.path.join(folder, "scheduledtasksinfo.csv")
        try:
            if os.path.exists(state_file):
                with open(file=state_file, mode="rb") as f:
                    state_output = f.read().decode(encoding="utf-16")
            if os.path.exists(details_file):
                with open(file=details_file, mode="rb") as f:
                    details_output = f.read().decode(encoding="utf-16")
        except Exception as e:
            logger.exception(e)
        if not state_output:
            logger.error(f"Failed to read results from {state_file}")
        if not details_output:
            logger.error(f"Failed to read results from {details_file}")

        return state_output, details_output